"""Keeps track of whether species are making progress and helps remove ones that are not."""
import sys

from .config import ConfigParameter, DefaultClassConfig
from .six_util import iteritems
from .math_util import stat_functions

# TODO: Add a method for the user to change the "is stagnant" computation.


class DefaultStagnation(DefaultClassConfig):
    """Keeps track of whether species are making progress and helps remove ones that are not."""
    @classmethod
    def parse_config(cls, param_dict):
        return DefaultClassConfig(param_dict,
                                  [ConfigParameter('species_fitness_func', str, 'mean'),
                                   ConfigParameter('max_stagnation', int, 15),
                                   ConfigParameter('species_elitism', int, 0)])

    def __init__(self, config, reporters):
        # pylint: disable=super-init-not-called
        self.stagnation_config = config

        self.species_fitness_func = stat_functions.get(config.species_fitness_func)
        if self.species_fitness_func is None:
            raise RuntimeError(
                "Unexpected species fitness func: {0!r}".format(config.species_fitness_func))

        self.reporters = reporters

    def update(self, species_set, generation):
        """
        Required interface method. Updates species fitness history information,
        checking for ones that have not improved in max_stagnation generations,
        and - unless it would result in the number of species dropping below the configured
        species_elitism parameter if they were removed,
        in which case the highest-fitness species are spared -
        returns a list with stagnant species marked for removal.
        """
        species_data = []
        for sid, s in iteritems(species_set.species):  # dla każdego gatunku
            if s.fitness_history:  # jeżeli lista pamietająca fitness jest pusta
                prev_fitness = max(s.fitness_history)  # weź max fitness
            else:
                # jeżeli jest pusta to weź możliwie najmniejszy fitness
                prev_fitness = -sys.float_info.max
            # oblicz fitness gatunku według funkcji podanej w config,
            # get_fitnessess zwraca listę fitnesów wszystkich osobnikow w danym gatunku
            s.fitness = self.species_fitness_func(s.get_fitnesses())
            s.fitness_history.append(s.fitness)  # dodaj obliczony fitness do historii
            s.adjusted_fitness = None
            # jeżeli max fitnes z historii nie istnieje lub aktualny jest lepszy
            if prev_fitness is None or s.fitness > prev_fitness:
                # zapisz numer genracji w której wystąpiła poprawa
                s.last_improved = generation

            species_data.append((sid, s))  # dodaj gatunek do pomocniczej listy gatunków

        # Sort in ascending fitness order.
        # posortuj rosnąco listę pomocniczą gatunków według fintessów gatunków
        species_data.sort(key=lambda x: x[1].fitness)

        result = []  # lista gatubnków do usunięcia
        species_fitnesses = []
        num_non_stagnant = len(species_data)  # ilosc gatunków
        for idx, (sid, s) in enumerate(species_data):  # dla każdego gatunku
            # Override stagnant state if marking this species as stagnant would
            # result in the total number of species dropping below the limit.
            # Because species are in ascending fitness order, less fit species
            # will be marked as stagnant first.
            stagnant_time = generation - s.last_improved  # ile generacji nie było poprawy
            is_stagnant = False  # flaga stagnacji
            # jezeli liczba wszystkich gatunkow jest większa niz liczba gatunków
            # podlegająca elityzmowi okreslona w config
            if num_non_stagnant > self.stagnation_config.species_elitism:
                # jeżeli gatunek nie poprawił się przez większą liczbę generacji
                # niż okreslona w config to podnies flage stagnacji dla danego gatunku
                is_stagnant = stagnant_time >= self.stagnation_config.max_stagnation

            # jezeli liczba pozostałych gautnkow do sprawdzenia jest mniejsza lub rowna
            # od liczby gatunkow podelgajacych elityzmowi, to spusc flage stagnacji
            # PAMIĘTAJ ŻE GATUNKI SĄ USZEREGOWANE ROSNĄCO PO FUNKCJI PRZYSTOSOWANIA
            # czyli jeżeli elityzmowi ma podlegać n gatunkow,
            # to dla ostatnich n gatunkow zgas flage
            # nawet jezeli nie poprawily się przed dlugi czas
            if (len(species_data) - idx) <= self.stagnation_config.species_elitism:
                is_stagnant = False
            # jezeli gatunek podlega stagnacji zmniejsz liczbe gatunkow
            # nie podlegajacych stagnacji
            if is_stagnant:
                num_non_stagnant -= 1
            # zapisz w wyniku id gatunku, gatunek i flage stagnacji
            result.append((sid, s, is_stagnant))
            species_fitnesses.append(s.fitness)  # zapisz fitness do  gatunku

        return result
